#include "burp.h"
#include "prog.h"
#include "asyncio.h"
#include "librsync.h"
#include "findlib/find.h"

static int fd=-1;

static char *readbuf=NULL;
static size_t readbuflen=0;
static size_t readbufmaxsize=(rs_inbuflen*2)+32;

static char *writebuf=NULL;
static size_t writebuflen=0;
static size_t writebufmaxsize=(rs_outbuflen*2)+32;

static void add_fd_to_sets(fd_set *read_set, fd_set *write_set, fd_set *err_set, int *max_fd)
{
	if(read_set) FD_SET((unsigned int) fd, read_set);
	if(write_set) FD_SET((unsigned int) fd, write_set);
	if(err_set) FD_SET((unsigned int) fd, err_set);

	if(fd > *max_fd) *max_fd = fd;
}

static void truncate_buf(char **buf, size_t *buflen)
{
	(*buf)[0]='\0';
	*buflen=0;
}

static int parse_readbuf(char *cmd, char **dest, size_t *rlen)
{
	unsigned int s=0;
	char cmdtmp='\0';

	if(readbuflen>=5)
	{
		if((sscanf(readbuf, "%c%04X", &cmdtmp, &s))!=2)
		{
			logp("sscanf of '%s' failed in parse_readbuf\n",
				readbuf);
			truncate_buf(&readbuf, &readbuflen);
			return -1;
		}
	}
	if(readbuflen>=s+5)
	{
		*cmd=cmdtmp;
		if(!(*dest=(char *)malloc(s+1)))
		{
			logp("out of memory in parse_readbuf\n");
			truncate_buf(&readbuf, &readbuflen);
			return -1;
		}
		if(!(memcpy(*dest, readbuf+5, s)))
		{
			logp("memcpy failed in parse_readbuf\n");
			truncate_buf(&readbuf, &readbuflen);
			return -1;
		}
		(*dest)[s]='\0';
		if(!(memmove(readbuf, readbuf+s+5, readbuflen-s-5)))
		{
			logp("memmove failed in parse_readbuf\n");
			truncate_buf(&readbuf, &readbuflen);
			return -1;
		}
		readbuflen-=s+5;
		*rlen=s;
	}
	return 0;
}

static int async_alloc_buf(char **buf, size_t *buflen, size_t bufmaxsize)
{
	if(!*buf)
	{
		if(!(*buf=(char *)malloc(bufmaxsize)))
		{
			logp("out of memory in async_alloc_buf\n");
			return -1;
		}
		truncate_buf(buf, buflen);
	}
	return 0;
}

static int do_read(void)
{
	ssize_t r;
#ifdef HAVE_WIN32
	if((r=recv(fd, readbuf+readbuflen, readbufmaxsize-readbuflen, 0))<=0)
#else
	if((r=read(fd, readbuf+readbuflen, readbufmaxsize-readbuflen))<=0)
#endif
	{
	//	logp("read returned: %d\n", r);
		truncate_buf(&readbuf, &readbuflen);
		return -1;
	}
	//logp("read: %d\n", r);
	readbuflen+=r;
	readbuf[readbuflen]='\0';
	return 0;
}

static int do_write(void)
{
	ssize_t w;
#ifdef HAVE_WIN32
	if((w=send(fd, writebuf, writebuflen, 0))<=0)
#else
	if((w=write(fd, writebuf, writebuflen))<=0)
#endif
	{
		logp("write returned: %d\n", w);
		return -1;
	}
	//logp("wrote: %d\n", w);
	memmove(writebuf, writebuf+w, writebuflen-w);
	writebuflen-=w;
	return 0;
}

static int append_to_write_buffer(const char *buf, size_t len)
{
	memcpy(writebuf+writebuflen, buf, len);
	writebuflen+=len;
	writebuf[writebuflen]='\0';
	return 0;
}

int async_append_all_to_write_buffer(char wcmd, const char *wsrc, size_t *wlen)
{
	size_t sblen=0;
	char sbuf[10]="";
	if(writebuflen+6+(*wlen) >= writebufmaxsize-1)
		return 1;

	snprintf(sbuf, sizeof(sbuf), "%c%04X", wcmd, (unsigned int)*wlen);
	sblen=strlen(sbuf);
	append_to_write_buffer(sbuf, sblen);
	append_to_write_buffer(wsrc, *wlen);
	//logp("appended to wbuf: %c (%d) (%d)\n", wcmd, *wlen+sblen, writebuflen);
	*wlen=0;
	return 0;
}

int async_init(int afd)
{
	fd=afd;
	if(async_alloc_buf(&readbuf, &readbuflen, readbufmaxsize)
	  || async_alloc_buf(&writebuf, &writebuflen, writebufmaxsize))
		return -1;
	return 0;

}

void async_free(void)
{
	close_fd(&fd);
	readbuflen=0;
	writebuflen=0;
	if(readbuf) { free(readbuf); readbuf=NULL; }
	if(writebuf) { free(writebuf); writebuf=NULL; }
}

/* for debug purposes */
static int setsec=1;
static int setusec=1;

void settimers(int sec, int usec)
{
	setsec=sec;
	setusec=usec;
}

int async_rw(char *rcmd, char **rdst, size_t *rlen, char wcmd, const char *wsrc, size_t *wlen)
{
        int mfd=-1;
        fd_set fsr;
        fd_set fsw;
        fd_set fse;
	int doread=0;
	int dowrite=0;
        struct timeval tval;
	if(fd<0)
	{
		logp("fd not ready in async rw: %d\n", fd);
		return -1;
	}

	if(rdst) doread++; // Given a pointer to allocate and read into.

	if(*wlen)
	{
		// More stuff to append to the write buffer.
		async_append_all_to_write_buffer(wcmd, wsrc, wlen);
	}

	if(writebuflen) dowrite++; // The write buffer is not yet empty.

	if(doread)
	{
		if(parse_readbuf(rcmd, rdst, rlen))
		{
			logp("error in parse_readbuf\n");
			return -1;
		}
		if(*rcmd && *rdst) return 0;
	}

        if(doread || dowrite)
        {
//logp("async_rw loop read %d write %d wbuflen: %d\n", doread, dowrite, writebuflen);
                mfd=-1;

                if(doread) FD_ZERO(&fsr);
                if(dowrite) FD_ZERO(&fsw);
                FD_ZERO(&fse);

                add_fd_to_sets(doread?&fsr:NULL, dowrite?&fsw:NULL, &fse, &mfd);

                tval.tv_sec=setsec;
                tval.tv_usec=setusec;

                if(select(mfd+1,
			doread?&fsr:NULL, dowrite?&fsw:NULL, &fse, &tval)<0)
                {
                        if(errno!=EAGAIN && errno!=EINTR)
                        {
                                logp("select error: %s\n", strerror(errno));
                                return -1;
                        }
                }

		if(!FD_ISSET(fd, &fse)
		  && (!doread || !FD_ISSET(fd, &fsr))
		  && (!dowrite || !FD_ISSET(fd, &fsw)))
		{
	//		logp("SELECT HIT TIMEOUT - doread: %d, dowrite: %d\n",
	//			doread, dowrite);
		}

                if(FD_ISSET(fd, &fse))
                {
                        logp("error on socket\n");
                        return -1;
                }

                if(doread && FD_ISSET(fd, &fsr)) // able to read
                {
			int r;
			if(do_read()) return -1;
			if((r=parse_readbuf(rcmd, rdst, rlen)))
				logp("error in second parse_readbuf\n");
			return r;
                }

                if(dowrite && FD_ISSET(fd, &fsw)) // able to write
		{
			int r;
			if((r=do_write()))
				logp("error in do_write\n");
			return r;
		}
        }

        return 0;
}

int async_rw_ensure_read(char *rcmd, char **rdst, size_t *rlen, char wcmd, const char *wsrc, size_t wlen)
{
	size_t w=wlen;
	while(!*rdst) if(async_rw(rcmd, rdst, rlen, wcmd, wsrc, &w))
		return -1;
	return 0;
}

int async_rw_ensure_write(char *rcmd, char **rdst, size_t *rlen, char wcmd, const char *wsrc, size_t wlen)
{
	size_t w=wlen;
	while(w) if(async_rw(rcmd, rdst, rlen, wcmd, wsrc, &w))
		return -1;
	return 0;
}

int async_read(char *rcmd, char **rdst, size_t *rlen)
{
	return async_rw_ensure_read(rcmd, rdst, rlen, '\0', NULL, 0);
}

int async_write(char wcmd, const char *wsrc, size_t wlen)
{
	return async_rw_ensure_write(NULL, NULL, NULL, wcmd, wsrc, wlen);
}

int async_write_str(char wcmd, const char *wsrc)
{
	size_t w;
	w=strlen(wsrc);
	return async_write(wcmd, wsrc, w);
}

int async_read_expect(char cmd, const char *expect)
{
	int ret=0;
	char rcmd='x';
	char *rdst=NULL;
	size_t rlen=0;
	if(async_read(&rcmd, &rdst, &rlen)) return -1;
	if(rcmd!=cmd || strcmp(rdst, expect))
	{
		logp("expected '%c:%s', got '%c:%s'\n",
			cmd, expect, rcmd, rdst);
		ret=-1;
	}
	free(rdst);
	return ret;
}


static int async_read_fp_msg(FILE *fp, gzFile zp, char **buf, size_t len)
{
	char *b=NULL;
	ssize_t r=0;

	/* Now we know how long the data is, so read it. */
	if(!(*buf=(char *)malloc(len+1)))
	{
		logp("could not malloc %d\n", len+1);
		return -1;
	}

	b=*buf;
	while(len>0)
	{
		if((zp && (r=gzread(zp, b, len))<=0)
		  || (fp && (r=fread(b, 1, len, fp))<=0))
		{
			//logp("read returned: %d\n", r);
			if(*buf) free(*buf);
			*buf=NULL;
			return -1;
		}
		b+=r;
		len-=r;
	}
	*b='\0';
	//logp("read_msg: %s\n", *buf);
	return 0;
}

int async_read_fp(FILE *fp, gzFile zp, char *cmd, char **rdst, size_t *rlen)
{
	unsigned int r;
	char *tmp=NULL;

	// First, get the command and length
	if(async_read_fp_msg(fp, zp, &tmp, 5))
	{
		if(tmp) free(tmp);
		return -1;
	}

	if((sscanf(tmp, "%c%04X", cmd, &r))!=2)
	{
		logp("sscanf of '%s' failed\n", tmp);
		if(tmp) free(tmp);
		return -1;
	}
	*rlen=r;
	if(tmp) free(tmp);

	return async_read_fp_msg(fp, zp, rdst, (*rlen)+1); // +1 for '\n'.
}

void log_and_send(const char *msg)
{
	logp("%s\n", msg);
	if(fd>0) async_write_str('e', msg);
}

int async_get_fd(void)
{
	return fd;
}



/* Read from fp if given - else read from our fd */
int async_read_stat(FILE *fp, gzFile zp, char **buf, size_t *len, struct stat *statp, char **dpth)
{
	char cmd;
	char *d=NULL;
	while(1)
	{
		if(fp || zp)
		{
			if(async_read_fp(fp, zp, &cmd, buf, len))
				break;
			if((*buf)[*len]=='\n') (*buf)[*len]='\0';
		}
		else
		{
			if(async_read(&cmd, buf, len))
				break;
		}
		if(cmd=='t')
		{
			if(d) free(d);
			d=*buf;
			*buf=NULL;
		}
		else if(cmd=='r')
		{
			if(statp) decode_stat(*buf, statp);
			if(*dpth) free(*dpth);
			*dpth=d;
			return 0;
		}
		else if((cmd=='c' && !strcmp(*buf, "backupend"))
		  || (cmd=='c' && !strcmp(*buf, "restoreend")))
		{
			if(d) free(d);
			return 1;
		}
		else
		{
			logp("expected cmd 't' or 'r', got '%c:%s'\n", cmd, *buf?:"");
			if(*buf) { free(*buf); *buf=NULL; }
			break;
		}
	}
	if(d) free(d);
	return -1;
}
