#include "burp.h"
#include "handy.h"
#include "prog.h"
#include "msg.h"
#include "asyncio.h"
#include "findlib/find.h"

void close_fp(FILE **fp)
{
	if(!*fp) return;
	fclose(*fp);
	*fp=NULL;
}

void gzclose_fp(gzFile *fp)
{
	if(!*fp) return;
	gzflush(*fp, Z_FINISH);
	gzclose(*fp);
	*fp=NULL;
}

int is_dir(const char *path)
{
        struct stat buf;

        if(lstat(path, &buf)) return 0;

        return S_ISDIR(buf.st_mode);
}

static char *prepend(const char *prep, const char *fname, size_t len, bool slash)
{
	char *rpath=NULL;

	if(prep) len+=strlen(prep)+1;
	if(slash) len++;
	len++;

	if(!(rpath=(char *)malloc(len)))
	{
		logp("could not malloc for relative path of %s\n", rpath);
		return NULL;
	}
	snprintf(rpath, len, "%s%s%s",
		prep?:"", (slash && *fname)?"/":"", fname);
	return rpath;
}

char *prepend_s(const char *prep, const char *fname, size_t len)
{
	if(!prep || !*prep)
	{
		char *ret=NULL;
		if(!(ret=strdup(fname)))
			logp("out of memory in prepend_s\n");
		return ret;
	}
	// Try to avoid getting a double slash in the path.
	if(fname && fname[0]=='/')
	{
		fname++;
		len--;
	}
	return prepend(prep, fname, len, TRUE);
}

int mkpath(char **rpath)
{
	char *cp=NULL;
	struct stat buf;
	//printf("mkpath: %s\n", *rpath);
	if((cp=strrchr(*rpath, '/')))
	{
		*cp='\0';
		if(lstat(*rpath, &buf))
		{
			// does not exist - recurse further down, then come
			// back and try to mkdir it.
			if(mkpath(rpath)) return -1;
			if(mkdir(*rpath, 0777))
			{
				logp("could not mkdir %s: %s\n", *rpath, strerror(errno));
				*cp='/';
				return -1;
			}
		}
		else if(S_ISDIR(buf.st_mode))
		{
			// Is a directory - can put the slash back and return.
		}
		else if(S_ISLNK(buf.st_mode))
		{
			// to help with the 'current' symlink
		}
		else
		{
			// something funny going on
			logp("warning: wanted '%s' to be a directory\n",
			*rpath);
		}
		*cp='/';
	}
	return 0;
}

int build_path(const char *datadir, const char *fname, size_t flen, char **rpath)
{
	if(!(*rpath=prepend_s(datadir, fname, flen))) return -1;
	if(mkpath(rpath))
	{
		if(*rpath) free(*rpath);
		return -1;
	}
	return 0;
}

#define ZCHUNK 16000

int send_whole_file_gz(const char *fname)
{
	int ret=0;
	int zret=0;
#ifdef HAVE_WIN32
	BFILE bfd;
#else
	FILE *fp=NULL;
#endif

	unsigned have;
	z_stream strm;
	int flush=Z_NO_FLUSH;
	unsigned char in[ZCHUNK];
	unsigned char out[ZCHUNK];

logp("send_whole_file_gz: %s\n", fname);

#ifdef HAVE_WIN32
	binit(&bfd);
	if(bopen(&bfd, fname, O_RDONLY | O_BINARY | O_NOATIME, 0)<0)
	{
		berrno be;
		logp("Could not open %s: %s\n", fname, be.bstrerror(errno));
		return -1;
	}
#else
	if(!(fp=fopen(fname, "rb")))
	{
		logp("Could not open %s: %s\n", fname, strerror(errno));
		return -1;
	}
#endif

	/* allocate deflate state */
	strm.zalloc = Z_NULL;
	strm.zfree = Z_NULL;
	strm.opaque = Z_NULL;
	if((zret=deflateInit2(&strm, Z_BEST_COMPRESSION, Z_DEFLATED, (15+16),
		8, Z_DEFAULT_STRATEGY))!=Z_OK)

	{
#ifdef HAVE_WIN32
		bclose(&bfd);
#else
		fclose(fp);
#endif
		return -1;
	}

	do
	{
#ifdef HAVE_WIN32
		strm.avail_in=(uint32_t)bread(&bfd, in, ZCHUNK);
#else
		strm.avail_in=fread(in, 1, ZCHUNK, fp);
#endif
		if(strm.avail_in<0)
		{
			logp("Error in read: %d\n", strm.avail_in);
			deflateEnd(&strm);
			ret=-1;
			break;
		}

		if(strm.avail_in) flush=Z_NO_FLUSH;
		else flush=Z_FINISH;

		strm.next_in=in;

		/* run deflate() on input until output buffer not full, finish
			compression if all of source has been read in */
		do
		{
			strm.avail_out = ZCHUNK;
			strm.next_out = out;
			zret = deflate(&strm, flush); /* no bad return value */
			if(zret==Z_STREAM_ERROR) /* state not clobbered */
			{
				logp("z_stream_error\n");
				ret=-1;
				break;
			}
			have = ZCHUNK-strm.avail_out;
			if(async_write('a', (const char *)out, have))
			{
				deflateEnd(&strm);
				ret=-1;
				break;
			}
		} while (!strm.avail_out);

		if(ret) break;

		if(strm.avail_in) /* all input will be used */
		{
			ret=-1;
			logp("strm.avail_in=%d\n", strm.avail_in);
			break;
		}
	} while(flush!=Z_FINISH);

	if(!ret && zret!=Z_STREAM_END)
	{
		logp("ret OK, but zstream not finished: %d\n", zret);
		ret=-1;
	}
	deflateEnd(&strm);

#ifdef HAVE_WIN32
	bclose(&bfd);
#else
	fclose(fp);
#endif
	return ret;
}

int send_whole_file(const char *fname)
{
	int ret=0;
	ssize_t s=0;
	char buf[4096]="";
#ifdef HAVE_WIN32
	BFILE bfd;
	binit(&bfd);
	if(bopen(&bfd, fname, O_RDONLY | O_BINARY | O_NOATIME, 0)<0)
	{
		berrno be;
		logp("Could not open %s: %s\n", fname, be.bstrerror(errno));
		return -1;
	}
	while((s=(uint32_t)bread(&bfd, buf, 4096))>0)
	{
		if(async_write('a', buf, s))
		{
			ret=-1;
			break;
		}
	}
	bclose(&bfd);
#else
	FILE *fp=NULL;
printf("send_whole_file: %s\n", fname);
	if(!(fp=fopen(fname, "rb")))
	{
		logp("Could not open %s: %s\n", fname, strerror(errno));
		return -1;
	}
	while((s=fread(buf, 1, 4096, fp))>0)
	{
		if(async_write('a', buf, s))
		{
			ret=-1;
			break;
		}
	}
	fclose(fp);
#endif
	return ret;
}

int set_non_blocking(int fd)
{
    int flags;
    if((flags = fcntl(fd, F_GETFL, 0))<0) flags = 0;
    return fcntl(fd, F_SETFL, flags | O_NONBLOCK);
}
     
int set_blocking(int fd)
{
    int flags;
    if((flags = fcntl(fd, F_GETFL, 0))<0) flags = 0;
    return fcntl(fd, F_SETFL, flags | ~O_NONBLOCK);
}
