#include "burp.h"
#include "conf.h"

static int conf_error(const char *config_path, int line)
{
	logp("%s: parse error on line %d\n", config_path, line);
	return -1;
}

void init_config(struct config *conf)
{
	conf->mode=MODE_UNSET;
	conf->port=0;
	conf->keep=0;
	conf->hardlinked_archive=0;
	conf->working_dir_recovery_method=NULL;
	conf->backupdir=NULL;
	conf->clientconfdir=NULL;
	conf->cname=NULL;
	conf->directory=NULL;
	conf->lockfile=NULL;
	conf->password=NULL;
	conf->server=NULL;
	conf->sigpath=NULL;
	conf->manifest=NULL;
}

void free_config(struct config *conf)
{
	if(!conf) return;
	if(conf->backupdir) free(conf->backupdir);
	if(conf->clientconfdir) free(conf->clientconfdir);
	if(conf->cname) free(conf->cname);
	if(conf->directory) free(conf->directory);
	if(conf->lockfile) free(conf->lockfile);
	if(conf->password) free(conf->password);
	if(conf->server) free(conf->server);
	if(conf->sigpath) free(conf->sigpath);
	if(conf->manifest) free(conf->manifest);
	if(conf->working_dir_recovery_method)
		free(conf->working_dir_recovery_method);
	init_config(conf);
}

static int get_conf_val(const char *field, const char *value, const char *want, char **dest)
{
	if(!strcmp(field, want) && !(*dest=strdup(value)))
	{
		logp("could not strdup %s value: %s\n", field, value);
		return -1;
	}
	return 0;
}

static int get_pair(char *buf, char **field, char **value)
{
	char *cp=NULL;
	char *eq=NULL;

	// strip leading space
	for(cp=buf; *cp && isspace(*cp); cp++) { }
	if(!*cp || *cp=='#')
	{
		*field=NULL;
		*value=NULL;
		return 0;
	}
	*field=cp;
	if(!(eq=strchr(*field, '='))) return -1;
	*eq='\0';

	// strip white space from before the equals sign
	for(cp=eq-1; *cp && isspace(*cp); cp--) *cp='\0';
	// skip white space after the equals sign
	for(cp=eq+1; *cp && isspace(*cp); cp++) { }
	*value=cp;
	if((cp=strrchr(*value, '\n'))) *cp='\0';

	if(!*field || !**field || !*value || !**value) return -1;

	return 0;
}

int load_config(const char *config_path, struct config *conf, bool loadall)
{
	int r=0;
	int line=0;
	FILE *fp=NULL;
	char buf[256]="";

	//logp("in load_config\n");
	if(!(fp=fopen(config_path, "r")))
	{
		logp("could not open '%s' for reading.\n", config_path);
		return -1;
	}
	while((fgets(buf, sizeof(buf), fp)))
	{
		char *field=NULL;
		char *value=NULL;
		line++;

		if(get_pair(buf, &field, &value))
			return conf_error(config_path, line);
		if(!field || !value) continue;

		if(!strcmp(field, "mode"))
		{
			if(!strcmp(value, "server"))
				conf->mode=MODE_SERVER;
			else if(!strcmp(value, "client"))
				conf->mode=MODE_CLIENT;
			else
				return conf_error(config_path, line);
		}
		else if(!strcmp(field, "port"))
		{
			conf->port=atoi(value);
			if(conf->port<=0)
				return conf_error(config_path, line);
		}
		else if(!strcmp(field, "keep"))
		{
			conf->keep=atoi(value);
			if(!conf->keep)
				return conf_error(config_path, line);
		}
		else if(!strcmp(field, "hardlinked_archive"))
		{
			conf->hardlinked_archive=atoi(value);
		}
		else if(!strcmp(field, "working_dir_recovery_method"))
		{
			if(get_conf_val(field, value,
			  "working_dir_recovery_method",
			  &(conf->working_dir_recovery_method))) return -1;
			if(strcmp(conf->working_dir_recovery_method, "delete")
			  && strcmp(conf->working_dir_recovery_method, "use")
			  && strcmp(conf->working_dir_recovery_method, "merge"))
			{
			  logp("unknown working_dir_recovery_method: %s\n",
				conf->working_dir_recovery_method);
			}
		}
		else
		{
			if(get_conf_val(field, value,
			  "backupdir", &(conf->backupdir))) return -1;
			if(get_conf_val(field, value,
			  "clientconfdir", &(conf->clientconfdir))) return -1;
			if(get_conf_val(field, value,
			  "cname", &(conf->cname))) return -1;
			if(get_conf_val(field, value,
			  "directory", &(conf->directory))) return -1;
			if(get_conf_val(field, value,
			  "lockfile", &(conf->lockfile))) return -1;
			if(get_conf_val(field, value,
			  "password", &(conf->password))) return -1;
			if(get_conf_val(field, value,
			  "server", &(conf->server))) return -1;
			if(get_conf_val(field, value,
			  "sigpath", &(conf->sigpath))) return -1;
			if(get_conf_val(field, value,
			  "manifest", &(conf->manifest))) return -1;
		}
	}
	fclose(fp);

	if(!loadall) return 0;

	if(conf->port<=0)
	{
		logp("%s: port unset\n", config_path);
		r--;
	}
	// Let the caller check the 'keep' value.
	if(!conf->lockfile)
	{
		logp("%s: lockfile unset\n", config_path);
		r--;
	}

	switch(conf->mode)
	{
		case MODE_SERVER:
			if(!conf->directory)
			  { logp("%s: directory unset\n", config_path); r--; }
			if(!conf->clientconfdir)
			  { logp("%s: clientconfdir unset\n", config_path); r--; }
			if(!conf->working_dir_recovery_method)
			  { logp("%s: working_dir_recovery_method unset\n", config_path); r--; }
			break;
		case MODE_CLIENT:
			if(!conf->cname)
			  { logp("%s: client name unset\n", config_path); r--; }
			if(!conf->backupdir)
			  { logp("%s: backupdir unset\n", config_path); r--; }
			if(!conf->password)
			  { logp("%s: password unset\n", config_path); r--; }
			if(!conf->server)
			  { logp("%s: server unset\n", config_path); r--; }
			if(!conf->sigpath)
			  { logp("%s: sigpath unset\n", config_path); r--; }
			if(!conf->manifest)
			  { logp("%s: manifest unset\n", config_path); r--; }
			break;
		case MODE_UNSET:
		default:
		  logp("%s: mode unset - should be 'server' or 'client'\n",
			config_path);
		  r--;
		  break;
	}
	//if(!r) logp("ok\n");

	return r;
}

int set_client_global_config(struct config *conf, struct config *cconf)
{
	cconf->keep=conf->keep;
	cconf->hardlinked_archive=conf->hardlinked_archive;
	if(!(cconf->working_dir_recovery_method=
		strdup(conf->working_dir_recovery_method)))
	{
		logp("out of memory in set_client_global_config\n");
		return -1;
	}
	return 0;
}
